"""
.. module:: likelihood

:Synopsis: Prototype likelihood and likelihood manager
:Author: Jesus Torrado

This module defines the main :class:`Likelihood` class, from which every likelihood
inherits, and the :class:`LikelihoodCollection` class, which groups and manages all the
individual likelihoods and is the actual instance passed to the sampler. Additionally, it
defines an MPI wrapper for the last one, that initialises the likelihoods once and passes
a copy to each MPI slave (rank>1) process.

Input: specifying likelihoods to explore
----------------------------------------

Likelihoods are specified under the `likelihood` block, together with their options:

.. code-block:: yaml

   likelihood:
     [likelihood 1]:
        [likelihood 1 option 1]: [value 1]
        [...]
     [likelihood 2]:
        [likelihood 2 option 2]: [value 2]
        [...]

Likelihood parameters are specified within the ``params`` block, as explained in
:doc:`params_prior`.

Code conventions and defauls
----------------------------

Each likelihood *lives* in a folder with its name under the ``likelihoods`` folder of the
source tree. In that folder, there must be at least *three* files:

- A trivial ``__init__.py`` file containing a single line: ``from [name] import [name]``,
  where ``name`` is the name of the likelihood, and it's folder.
- A ``name.py`` file, containing the particular class definition of the likelihood, 
  inheriting from the :class:`Likelihood` class (see below).
- A ``defaults.yaml`` containing a block:

  .. code-block:: yaml

     likelihood:
       [name]:
         [option 1]: [value 1]
         [...]
       params:
         [param 1]:
           prior:
             [prior info]
           [label, ref, etc.]

  The options and parameters defined in this file are the only ones recognised by the
  likelihood, and they are loaded automatically with their default values (options) and
  priors (parameters) by simply mentioning the likelihood in the input file, where one can
  re-define any of those options with a different value. The same parameter may be
  defined by different likelihoods -- in those cases, it needs to have the same default 
  information (prior, label, etc.) in the defaults file of those likelihoods.

.. note::

   Some *mock* likelihoods can have any number of non-predefined parameters, as long as
   they start with a certain prefix specified by the user with the option ``mock_prefix``
   of said likelihood.

.. note::

   Actually, there are some user-defined options that are common to all likelihoods and
   do not need to be specified in the ``defaults.yaml`` file, such as the computational
   ``speed`` of the likelihood (see :ref:`mcmc_speed_hierarchy`).


Creating your own likelihood
----------------------------

Since cobaya was created to be flexible, creating your own likelihood is very easy: simply
create a folder with its name under ``likelihoods`` in the source tree and follow the 
conventions explained above. Inside the class definition of your function, you can use any
of the attributes defined in the ``defaults.yaml`` file directly, and you only need to 
specify one, or at most three, functions
(see the :class:`Likelihood` class documentation below):

- A ``logp`` function taking a dictionary of (sampled) parameter values and returning a
  log-likelihood.
- An (optional) ``initialise`` function preparing any computation, importing any necessary
  code, etc.
- An (optional) ``close`` function doing whatever needs to be done at the end of the
  sampling (e.g. releasing memory).

You can use the :doc:`Gaussian likelihood <likelihood_gaussian>` as a guide.
If your likelihood needs a cosmological code, just define one in the input file and you
can automatically access it as an attribute of your class: ``[your_likelihood].theory``.
Use the :doc:`Planck likelihood <likelihood_planck>` as a guide to create your own
cosmological likelihood.

.. note:: ``theory`` and ``derived`` are reserved parameter names: you cannot use them!

"""
# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import sys
import traceback
from collections import OrderedDict as odict
from time import sleep
import numpy as np
from scipy import stats
import inspect
from itertools import chain

# Local
from cobaya.conventions import _external, _theory, _params
from cobaya.tools import get_class, get_external_function
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)

# Default options for all subclasses
class_options = {"speed": 1}


class Likelihood():
    """Likelihood class prototype."""

    # Generic initialisation -- do not touch
    def __init__(self, info, parametrisation, theory=None):
        # Load info of the likelihood
        for k in info:
            setattr(self, k, info[k])
        # Mock likelihoods: gather all parameters starting with `mock_prefix`
        if self.is_mock():
            all_params = (list(parametrisation.input_params()) +
                          list(parametrisation.output_params()))
            info[_params] = [p for p in all_params
                             if p.startswith(self.mock_prefix or "")]
        # Load parameters
        self.input_params = odict(
            [(p,p_info) for p,p_info in parametrisation.input_params().iteritems()
             if p in info[_params]])
        self.output_params = odict(
            [(p,p_info) for p,p_info in parametrisation.output_params().iteritems()
             if p in info[_params]])
        # Initialise
        self.theory = theory
        self.initialise()
    
    # What you *must* implement to create your own likelihood:

    # Optional
    def initialise(self):
        """Initialises the specifics of this likelihood."""
        pass

    # Mandatory
    def logp(self, **params_values):
        """
        Computes and returns the likelihood value.
        Takes as keyword arguments the parameters values.
        To get the derived parameters, pass a `derived` keyword with an empty dictionary.
        """
        return None

    # Optional
    def close(self):
        """Finalises the likelihood, if something needs to be done."""
        pass

    # What you *can* implement to create your own *mock* likelihood:

    def marginal(self, directions=None, params_values=None):
        """
        (For mock likelihoods only.)
        Computes the marginal likelihood.
        If nothing is specified, returns the total marginal likelihood.
        If some directions are specified (as a list, tuple or array), returns the marginal
        likelihood pdf over those directions evaluated at the given parameter values.
        """
        log.error("Exact marginal likelihood not defined.")
        raise HandledException

    # Other general methods

    def wait(self):
        if self.delay:
            log.debug("Sleeping for %f seconds.", self.delay)
        sleep(self.delay)

    def d(self):
        return len(self.input_params)

    def is_mock(self):
        return hasattr(self, "mock_prefix")


class LikelihoodExternalFunction(Likelihood):
    def __init__(self, name, info, theory=None):
        self.name = name
        self.theory = theory
        # Load info of the likelihood
        for k in info:
            setattr(self, k, info[k])
        # Store the external function and its arguments
        self.external_function = get_external_function(info[_external])
        argspec = inspect.getargspec(self.external_function)
        self.input_params = odict([(p, None) for p in argspec.args if p != "derived"])
        self.has_derived = "derived" in argspec.args
        if self.has_derived:
            derived_kw_index = argspec.args[-len(argspec.defaults):].index("derived")
            self.output_params = argspec.defaults[derived_kw_index]
        else:
            self.output_params = []

    def logp(self, **params_values):
        # if no derived params defined in the external call, delete the "derived" argument
        if not self.has_derived:
            params_values.pop("derived")
        try:
            return self.external_function(**params_values)
        except:
            log.error("".join(
                ["-"]*16 + ["\n\n"] + list(traceback.format_exception(*sys.exc_info())) +
                ["\n"] + ["-"]*37))
            log.error("The external likelihood '%s' failed at evaluation. "
                      "See traceback on top of this message.", self.name)
            raise HandledException


class LikelihoodCollection():
    """
    Likelihood manager:
    Initialises the theory code and the experimental likelihoods.
    """

    def __init__(self, info_likelihood, parametrisation, info_theory=None):
        # Store the input/output parameters
        self.input_params = parametrisation.input_params()
        self.output_params = parametrisation.output_params()
        # *IF* there is a theory code, initialise it
        if info_theory:
            input_params_theory = self.input_params.fromkeys(
                [k for k in self.input_params if k in parametrisation.theory_params()])
            output_params_theory = self.output_params.fromkeys(
                [k for k in self.output_params if k in parametrisation.theory_params()])
            name, fields = info_theory.items()[0]
            theory_class = get_class(name, kind=_theory)
            self.theory = theory_class(input_params_theory, output_params_theory, fields)
        else:
            self.theory = None
        # Initialise individual Likelihoods
        self._likelihoods = odict()
        for name, info in info_likelihood.iteritems():
            # If it does "external" key, wrap it up. Else, load it up
            if _external in info:
                self._likelihoods[name] = LikelihoodExternalFunction(
                    name, info, theory=getattr(self, _theory, None))
            else:
                lik_class = get_class(name)
                self._likelihoods[name] = lik_class(
                    info, parametrisation, theory=self.theory)
        # Check that all are recognised
        for params in ("input_params", "output_params"):
            info = getattr(parametrisation, params)()
            setattr(self, params, info)
            requested = set(info)
            known = set(chain(getattr(self.theory, params, []),
                              *[getattr(self[lik], params) for lik in self]))
            r_not_k = requested.difference(known)
            if r_not_k:
                log.error("Some of the requested %s parameters were not recognised "
                          "by any likelihood: %r.", params.split("_")[0], r_not_k)
                raise HandledException
        # Store the input params and likelihods on which each sampled params depends
        self.sampled_input_dependence = parametrisation.sampled_input_dependence()
        self.sampled_lik_dependence = odict(
            [[p,[lik for lik in list(self)+([_theory] if self.theory else [])
                 if any([(i in self[lik].input_params) for i in (i_s or [p])])]]
             for p,i_s in self.sampled_input_dependence.items()])
        # Pop the per-likelihood parameters info, that was auxiliary
        for lik in info_likelihood:
            info_likelihood[lik].pop(_params)

    # "get" and iteration operate over the dictionary of likelihoods
    # notice that "get" can get "theory", but the iterator does not!
    def __getitem__(self, key):
        return self._likelihoods.__getitem__(key) if key != _theory else self.theory

    def __iter__(self):
        return self._likelihoods.__iter__()

    def logps(self, input_params, derived=None):
        """
        Computes observables and returns the (log) likelihoods *separately*.
        It takes a list of **input** parameter values, in the same order as they appear
        in the `OrderedDictionary` of the :class:LikelihoodCollection.
        To compute the derived parameters, it takes an optional keyword `derived` as an
        empty list, which is then populated with the derived parameters values.
        """
        # Prepare the likelihood-defined derived parameters (only computed if requested)
        # Notice that they are in general different from the sampler-defined ones.
        derived_dict = {}
        # If theory code present, compute the necessary products
        if self.theory:
            this_params_dict = {p: input_params[p] for p in self.theory.input_params}
            success = self.theory.compute(
                derived=(derived_dict if derived is not None else None),
                **this_params_dict)
            if not success:
                if derived is not None:
                    derived += [np.nan]*len(self.output_params)
                return np.array([-np.inf for _ in self])
        # Compute each log-likelihood, and optionally get the respective derived params
        logps = []
        for lik in self:
            this_params_dict = {p: input_params[p] for p in self[lik].input_params}
            if derived is not None:
                this_derived_dict = {}
            logps += [self[lik].logp(derived=this_derived_dict, **this_params_dict)]
            derived_dict.update(this_derived_dict)
        # Turn the derived params dict into a list and return
        if derived is not None:
            derived += [derived_dict[p] for p in self.output_params]
        return np.array(logps)

    def logp(self, input_params, derived=None):
        """
        Computes observables and returns the (log) likelihood.
        It takes a list of **sampled** parameter values, in the same order as they appear
        in the `OrderedDictionary` of the :class:LikelihoodCollection.
        To compute the derived parameters, it takes an optional keyword `derived` as an
        empty list, which is then populated with the derived parameters values.
        """
        return np.sum(self.logps(input_params, derived=derived))

    def marginal(self, *args):
        log.error("Marginal not implemented for >1 likelihoods. Sorry!")
        raise HandledException

    def d(self):
        return sum([self[lik].d() for lik in self])

    def speeds_of_params(self):
        """
        Blocks the sampled parameters by likelihood, and sorts the blocks by speed.
        Returns an ``OrderedDict`` ``{speed: [params]}``, sorted by ascending speeds.
        Parameters recognised by more than one likelihood are blocked in the slowest one.
        """
        param_with_speed = odict([[p,min([self[lik].speed for lik in liks])]
                                  for p,liks in self.sampled_lik_dependence.items()])
        if 0 in param_with_speed.values():
            log.error("No likelihood can have 0 speed.")
            raise HandledException
        # Invert it!
        return odict([[speed,[p for p,speed2 in param_with_speed.items()
                              if speed == speed2]]
                      for speed in sorted(list(set(param_with_speed.values())))])

    # Python magic for the "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        for lik in self:
            self[lik].close()


# class LikelihoodCollection_MPI(LikelihoodCollection):
#     """
#     MPI wrapper around the LikelihoodCollection class.
#     """
#     def __init__(self, info_likelihood, info_params, info_theory=None):
#         from mpi4py import MPI
#         comm = MPI.COMM_WORLD
#         rank = comm.Get_rank()
#         to_broadcast = ("theory", "likelihoods", "fixed", "sampled", "derived",
#                         "_updated_info_params_liks")
#         if rank == 0:
#             LikelihoodCollection.__init__(
#                 self, info_likelihood, info_params, info_theory=info_theory)
#         else:
#             for var in to_broadcast:
#                 setattr(self, var, None)
#         for var in to_broadcast:
#             setattr(self, var, comm.bcast(getattr(self, var), root=0))



"""
.. module:: planck_clik_prototype

:Synopsis: Definition of the clik code based likelihoods
:Author: Julien Lesgourgues and Benjamin Audren (and Jesus Torrado for small compatibility changes only)

Contains the definition of the base clik 2.0 likelihoods, from which all Planck 2015
likelihoods inherit.

The treatment on Planck 2015 likelihoods has been adapted without much modification from
the `MontePython <http://baudren.github.io/montepython.html>`_ code by Julien Lesgourgues
and Benjamin Audren.

.. note::
  
  An older, `MIT-licensed <https://opensource.org/licenses/MIT>`_ version of this module by
  Julien Lesgourgues and Benjamin Audren can be found in the `source of
  MontePython <https://github.com/baudren/montepython_public>`_.


The Planck 2015 likelihoods defined here are:

- ``planck_2015_lowl``
- ``planck_2015_lowTEB``
- ``planck_2015_plikHM_TT``
- ``planck_2015_plikHM_TTTEEE``
- ``planck_2015_plikHM_TTTEEE_unbinned``
- ``planck_2015_plikHM_TT_unbinned``


Usage
-----

To use the Planck likelihoods, you simply need to mention them in the likelihood blocks,
specifying the path where you have installed them, for example:

.. code-block:: yaml

   likelihood:
     planck_2015_lowTEB:
       path: /path/to/cosmo/likelihoods/planck2015/
     planck_2015_plikHM_TTTEEE:
       path: /path/to/cosmo/likelihoods/planck2015/

This automatically pulls the likelihood parameters into the sampling, with their default
priors. The parameters can be fixed or their priors modified by re-defining them in the
``params: likelihood:`` block.

The default parameter and priors can be found in the ``defaults.yaml`` files in the
folder corresponding to each likelihood. They are not reproduced here because of their
length.

.. [The installation instructions are in the corresponding doc file because they need
..  some unicode characters to show some directory structure]

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
import sys
import numpy as np

# Local
from cobaya.likelihood import Likelihood
from cobaya.log import HandledException
from cobaya.conventions import _likelihood, subfolders, _path_install
from cobaya.tools import get_path_to_installation

# Logger
import logging
log = logging.getLogger(__name__)

class planck_clik_prototype(Likelihood):

    def initialise(self):
        self.name = self.__class__.__name__
        # Importing Planck's clik library (only once!)
        try:
            clik
        except NameError:
            if not self.path:
                path_to_installation = get_path_to_installation()
                if path_to_installation:
                    self.path = os.path.join(
                        path_to_installation, subfolders[_likelihood], common_path)
                else:
                    log.error("No path given to the Planck likelihood. Set the likelihood "
                              "property 'path' or the common property '%s'."
                              %_path_install)
                    raise HandledException
            log.info("[%s] Importing clik from %s", self.name, self.path)
            # test and import clik
            is_installed_clik(self.path, log_and_fail=True, import_it=False)
            import clik
        # Loading the likelihood data
        clik_file = os.path.join(self.path, self.clik_file)
        # for lensing, some routines change. Intializing a flag for easier
        # testing of this condition
        if 'lensing' in self.name and 'Planck' in self.name:
            self.lensing = True
        else:
            self.lensing = False
        try:
            if self.lensing:
                self.clik = clik.clik_lensing(clik_file)
                try: 
                    self.l_max = max(self.clik.get_lmax())
                # following 2 lines for compatibility with lensing likelihoods of 2013 and before
                # (then, clik.get_lmax() just returns an integer for lensing likelihoods;
                # this behavior was for clik versions < 10)
                except:
                    self.l_max = self.clik.get_lmax()
            else:
                self.clik = clik.clik(clik_file)
                self.l_max = max(self.clik.get_lmax())
        except clik.lkl.CError:
            log.error(
                "The path to the .clik file for the likelihood "
                "%s was not found where indicated."
                " Note that the default path to search for it is"
                " one directory above the path['clik'] field. You"
                " can change this behaviour in all the "
                "Planck_something.data, to reflect your local configuration, "
                "or alternatively, move your .clik files to this place.", self.name)
            raise HandledException
        except KeyError:
            log.error(
                "In the %s.data file, the field 'clik' of the "
                "path dictionary is expected to be defined. Please make sure"
                " it is the case in you configuration file", self.name)
            raise HandledException
        # Requested spectra
        if self.lensing:
            raise NotImplementedError("Lensing lik not implemented!!!")
            # For lensing, the order would be: **phiphi TT EE BB TE TB EB**
        requested_i = [c=="1" for c in self.clik.get_has_cl()]
        requested_cls =[cl for cl,i in
                        zip(["TT","EE","BB","TE","TB","EB"], requested_i) if i]
        # State requisites to the theory code
        self.theory.needs({"Cl": requested_cls, "l_max": self.l_max})
        self.expected_params = list(self.clik.extra_parameter_names)
        # line added to deal with a bug in planck likelihood release:
        # A_planck called A_Planck in plik_lite
        if "plikHM_lite" in self.name:
            i = self.expected_params.index('A_Planck')
            self.expected_params[i] = 'A_planck'
            log.info("In %s, corrected nuisance parameter name A_Planck to A_planck" % self.name)
        
    def logp(self, **params_values):
        # get Cl's from the theory code
        cl = self.theory.get_cl()
        # testing for lensing
        if self.lensing:
            try:
                length = len(self.clik.get_lmax())
                tot = np.zeros(
                    np.sum(self.clik.get_lmax()) + length +
                    len(self.params()))
            # following 3 lines for compatibility with lensing likelihoods of 2013 and before
            # (then, clik.get_lmax() just returns an integer for lensing likelihoods,
            # and the length is always 2 for cl['pp'], cl['tt'])
            except:
                length = 2
                tot = np.zeros(2*self.l_max+length + len(self.params()))
        else:
            length = len(self.clik.get_has_cl())
            tot = np.zeros(
                np.sum(self.clik.get_lmax()) + length +
                len(self.expected_params))
        # fill with Cl's
        index = 0
        if not self.lensing:
            for i in range(length):
                if (self.clik.get_lmax()[i] > -1):
                    for j in range(self.clik.get_lmax()[i]+1):
                        if (i == 0):
                            tot[index+j] = cl['tt'][j]
                        if (i == 1):
                            tot[index+j] = cl['ee'][j]
                        if (i == 2):
                            tot[index+j] = cl['bb'][j]
                        if (i == 3):
                            tot[index+j] = cl['te'][j]
                        if (i == 4):
                            tot[index+j] = 0 #cl['tb'][j] class does not compute tb
                        if (i == 5):
                            tot[index+j] = 0 #cl['eb'][j] class does not compute eb
                    index += self.clik.get_lmax()[i]+1
        else:
            try:
                for i in range(length):
                    if (self.clik.get_lmax()[i] > -1):
                        for j in range(self.clik.get_lmax()[i]+1):
                            if (i == 0):
                                tot[index+j] = cl['pp'][j]
                            if (i == 1):
                                tot[index+j] = cl['tt'][j]
                            if (i == 2):
                                tot[index+j] = cl['ee'][j]
                            if (i == 3):
                                tot[index+j] = cl['bb'][j]
                            if (i == 4):
                                tot[index+j] = cl['te'][j]
                            if (i == 5):
                                tot[index+j] = 0 #cl['tb'][j] class does not compute tb
                            if (i == 6):
                                tot[index+j] = 0 #cl['eb'][j] class does not compute eb

                        index += self.clik.get_lmax()[i]+1
            # following 8 lines for compatibility with lensing likelihoods of 2013 and before
            # (then, clik.get_lmax() just returns an integer for lensing likelihoods,
            # and the length is always 2 for cl['pp'], cl['tt'])
            except:
                for i in range(length):
                    for j in range(self.l_max):
                        if (i == 0):
                            tot[index+j] = cl['pp'][j]
                        if (i == 1):
                            tot[index+j] = cl['tt'][j]
                    index += self.l_max+1
        # fill with likelihood parameters
        for i,p in enumerate(self.expected_params):
            tot[index+i] = params_values[p]
        # In case there are derived parameters in the future:
        # derived = params_values.get("derived")
        # if derived != None:
        #     derived["whatever"] = [...]
        # Compute the likelihood
        return self.clik(tot)[0]

    def close(self):
        del(self.clik) # MANDATORY: forces deallocation of the Cython class
        # Actually, it does not work for low-l likelihoods, which is quite dangerous!


# Installation routines ###################################################################

# path to be shared by all Planck likelihoods
common_path = "planck_2015"

def download_from_planck(product_id, path):
    try:
        from wget import download, bar_thermometer
        wget_kwargs = {"out": path, "bar": bar_thermometer}
        prefix = r"http://pla.esac.esa.int/pla-sl/data-action?COSMOLOGY.COSMOLOGY_OID="
        filename = download(prefix+product_id, **wget_kwargs)
    except:
        log.error("Error downloading!")
        return False
    finally:
        print "" # force newline after wget
    # uncompress
    import os
    import tarfile
    extension = os.path.splitext(filename)[-1][1:]
    tar = tarfile.open(filename, "r:"+extension)
    try:
        tar.extractall(path)
        tar.close()
        os.remove(filename)
        return True
    except:
        log.error("Error decompressing downloaded file! Corrupt file?)")
        return False

def is_installed_clik(path, log_and_fail=False, import_it=True):
    clik_path = os.path.join(path, "plc-2.0")
    if not os.path.exists(clik_path):
        if log_and_fail:
            log.error("The given folder does not exist: '%s'",clik_path)
            raise HandledException
        return False
    clik_path = os.path.join(clik_path, "lib/python2.7/site-packages")
    if not os.path.exists(clik_path):
        if log_and_fail:
            log.error("You have not compiled the Planck likelihood code 'clik'.\n"
                      "Take a look at the docs to see how to do it using 'waf'.")
            raise HandledException
        return False
    sys.path.insert(0, clik_path)
    try:
        if import_it:
            import clik
        return True
    except:
        return False

def install_clik(path):
    log.info("clik: downlowading...")
    if not download_from_planck("1904", path):
        log.error("Not possible to download clik.")
        return False
    log.info("clik: configuring... (and maybe installing dependencies...)")
    os.chdir(os.path.join(path, "plc-2.0"))
    from subprocess import Popen, PIPE
    process = Popen(["./waf", "configure", "--install_all_deps"], stdout=PIPE, stderr=PIPE)
    out, err = process.communicate()
    if err != "" or not out.split("\n")[-2].startswith("'configure' finished successfully"):
        print out
        print err
        log.error("Configuration failed!")
        return False
    log.info("clik: compiling...")
    process2 = Popen(["./waf", "install"], stdout=PIPE, stderr=PIPE)
    out2, err2 = process2.communicate()
    # We don't check that err2" is empty, because harmless warnings are included there.
    if not out2.split("\n")[-2].startswith("'install' finished successfully"):
        print out2
        print err2
        log.error("Compilation failed!")
        return False
    log.info("clik: finished!")
    return True

def get_product_id_and_clik_file(name):
    # get it from the defaults.yaml file
    from cobaya.conventions import _defaults_file, _likelihood
    path__defaults_file = os.path.join(os.path.dirname(__file__), "..", name, _defaults_file)
    from cobaya.yaml_custom import yaml_load_file
    defaults = yaml_load_file(path__defaults_file)[_likelihood][name]
    return defaults["product_id"], defaults["clik_file"]

def is_installed(**kwargs):
    _, filename = get_product_id_and_clik_file(kwargs["name"])
    return os.path.exists(os.path.realpath(
        os.path.join(kwargs["path"], "..", common_path, filename)))

def install(force=False, path=None, name=None):
    # Create common folder: all planck likelihoods share install folder
    common_full_path = os.path.realpath(os.path.join(path, "..", common_path))
    if not os.path.exists(common_full_path):
        os.makedirs(common_full_path)
    # Install clik
    if not is_installed_clik(common_full_path) or force:
        log.info("Installing the clik code first!")
        success = install_clik(common_full_path)
        if not success:
            return False
    # Extract product_id
    product_id, _ = get_product_id_and_clik_file(name)
    # Download and uncompress the particular likelihood
    log.info("Downloading likelihood data...")
    if not download_from_planck(product_id, common_full_path):
        log.error("Not possible to download this likelihood.")
        return False
    log.info("Likelihood data downloaded and uncompressed correctly.")
    return True



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
# Package data and conventions
from cobaya import __author__, __version__, __url__
from cobaya.conventions import subfolders, _defaults_file

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cobaya',
    version=__version__,
    description='Bayesian Analysis in Cosmology',
    long_description=long_description,
    url=__url__,
    author=__author__,
    license='LGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='montecarlo sampling cosmology',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=['numpy>=1.7.1', 'scipy >= 0.18', 'pandas>=0.17.1',
                      'PyYAML>=3.12', 'wget>=3.2'],
    python_requires='>=2.7, <3',
    package_data={
        'cobaya': ['%s/*/%s'%(folder, _defaults_file) for folder in subfolders.values()]},
    entry_points={
        'console_scripts': [
            'cobaya-install=cobaya.install:install_script',
            'cobaya-run=cobaya.run:run_script',
            'cobaya-citation=cobaya.citation:citation_script',
        ],
    },
)



"""
.. module:: theory

:Synopsis: Prototype theory class and theory loader
:Author: Jesus Torrado

If you are using a experimental likelihood, chances are the you will need a theoretical
code to compute the observables needed to compute the likelihood.

This module contains the prototype of the theory code and the loader for the requested
code.

.. note::

   At this moment, of all modules of cobaya, this is the one with the least fixed
   structure. Don't pay much attention to it for now. Just go on to the installation
   instructions for :doc:`CAMB <theory_camb>` and :doc:`CLASS <theory_class>`.

"""

# Python 2/3 compatibility
from __future__ import division

# Global
import sys
import os
from collections import OrderedDict as odict
from importlib import import_module

# Local
from cobaya.conventions import _theory
from cobaya.log import HandledException

# Default options for all subclasses
class_options = {"speed": 1}


# Theory code prototype
class Theory():
    """Prototype of the theory class."""
    
    def initialise(self):
        """
        Initialises the theory code: imports the theory code, if it is an external one, 
        and makes any necessary preparations.
        """
        pass

    def needs(self, arguments):
        """
        Function to be called by the likelihoods at their initialisation,
        to specify their requests.
        Its specific behaviour for a code must be defined.
        """
        pass

    def compute(self, **parameter_values_and_derived_dict):
        """
        Takes a dictionary of parameter values and computes the products needed by the
        likelihood.
        If passed a keyword `derived` with an empty dictionary, it populates it with the
        value of the derived parameters for the present set of sampled and fixed parameter
        values.
        """
        pass
    
    def close(self):
        """Finalises the theory code, if something needs to be done
        (releasing memory, etc.)"""
        pass

    # Generic methods: do not touch these

    def __init__(self, info_input_params, info_output_params, info_theory):
        self.input_params = info_input_params
        self.output_params = info_output_params
        # Load info of the code
        for k in info_theory:
            setattr(self, k, info_theory[k])
        self.initialise()

    def d(self):
        return len(self.sampled)



"""
.. module:: citation

:Synopsis: Tools and script to get the references to cite
:Author: Jesus Torrado

Inspired by a similar characteristic of
`CosmoSIS <https://bitbucket.org/joezuntz/cosmosis/wiki/Home>`_.

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os

# Local
from cobaya.tools import get_folder, make_header
from cobaya.input import get_modules

def get_citation_info(module, kind):
    folder = get_folder(module, kind, absolute=True)
    filename = os.path.join(folder, module+".bibtex")
    try:
        with open(filename, "r") as f:
            lines = "".join(f.readlines())
    except IOError:
        if not os.path.isdir(folder):
            lines = "[Module '%s.%s' not known.]"%(kind, module)
        else:
            lines = "[no citation information found]"
    return lines+"\n"

def citation(*infos):
    print make_header("This framework", "")
    print "[Paper in preparation]\n"
    for kind, modules in get_modules(*infos).iteritems():
        for module in modules:
            print make_header(kind, module)
            print get_citation_info(module, kind)

# Command-line script
def citation_script():
    from cobaya.mpi import get_mpi_rank
    if not get_mpi_rank():
        # Configure the logger ASAP
        from cobaya.log import logger_setup
        logger_setup()
        # Parse arguments and launch
        import argparse
        parser = argparse.ArgumentParser(description="Cobaya's citation tool.")
        parser.add_argument("files", action="store", nargs="+", metavar="input_file.yaml",
                            help="One or more input files.")
        from cobaya.input import load_input
        infos = [load_input(f) for f in parser.parse_args().files]
        citation(*infos)



"""
.. module:: prior

:Synopsis: Class containing the prior and reference pdf, and other parameter information.
:Author: Jesus Torrado

The ``params`` block contains all the information intrinsic
to the parameters of the model: their prior pdf and reference pdf, their latex labels,
or useful properties for particular samplers (e.g. with of a proposal pdf in and MCMC
-- see the docs section for each sampler). All this information is stored and managed by
the :class:`Prior` class.

You can specify three different kinds of parameters:

+ **Fixed** parameters are specified by assigning them a value, and are passed directly to
  the likelihood or theory code.
+ **Sampled** parameters need a ``prior`` pdf definition and, optionally,
  a reference pdf or value (``ref``), a LaTeX label to be used when producing plots,
  and additional properties to aid particular samplers.
+ **Derived** parameters do **not** have a prior definition (neither a reference value or pdf),
  but can have a LaTeX label and a ``min`` and/or ``max`` to be used in the sample analysis
  (e.g. to guess the pdf tails correctly if the derived quantity needs to be positive or
  negative -- defaulted to ``-inf``, ``inf`` resp.).

The (optional) **reference** pdf (``ref``) defines the region of the prior which is of most
interest (e.g. where most of the prior mass is expected); samplers needing an initial point
for a chain will attempt to draw it from the ``ref`` if it has been defined (otherwise
from the prior). A good reference pdf will avoid a long *burn-in* stage during the
sampling.

The syntax for priors and ref's has three fields:

+ ``dist`` (default: ``uniform``): any 1-dimensional continuous distribution from
  `scipy.stats <https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions>`_.
+ ``loc`` and ``scale`` (default: 0 and 1, resp.): the *location* and *scale* of the pdf,
  as they are understood for each particular pdf in :class:`scipy.stats` (e.g. for a
  ``uniform`` pdf, ``loc`` is the lower bound and ``scale`` is the length of the domain,
  whereas in a Gaussian ``norm`` pdf ``loc`` is the centre and ``scale`` is the standard
  deviation).
+ (alternatively) ``min`` and ``max`` (default: 0 and 1 resp.): the boundaries of the pdf
  for a bounded distribution (``uniform``, ``beta``, etc. -- unexpected behaviour for an
  unbounded one).

The order of the parameters is conserved in the table of samples, except the fact that
derived parameters are always moved to the last places.

An example ``params`` block:

.. code-block :: yaml

     params:

       A: 1 # fixed!

       B1:  # sampled! with uniform prior on [-1,1] and ref pdf N(mu=0,sigma=0.25)
         prior:
           min: -1
           max:  1
         ref:
           dist: norm
           loc:   0
           scale: 0.25
         latex: \mathcal{B}_2
         proposal: 0.25

       B2:  # sampled! with prior N(mu=0.5,sigma=3) and fixed reference value 1
         prior:
           dist: norm
           loc: 0.5
           scale: 3
         ref: 1  # fixed reference value: all chains start here!
         latex: \mathcal{B}_2
         proposal: 0.5

       C:  # derived!
         min: 0
         latex: \mathcal{C}

.. note:: ``theory`` is a reserved parameter name!

.. warning::

   Improper priors will produce unexpected errors. So don't try

   .. code-block:: yaml

      params:
        a:
          prior:
            min: -inf
            max:  inf

Changing and redefining parameters -- inheritance
-------------------------------------------------

When mentioning a likelihood, the parameters that appear in its ``defaults.yaml`` file are
inherited without needing to re-state them. Still, you may want to re-define their type or
some of their properties in your input file:

* Re-defining **fixed** into:

 - **fixed**: simply assign it a new value
 - **sampled**: define a prior for it (and optionally reference, label, etc.)
 - **derived**: mention the parameter and assign nothing to it (you may define a
   min, max or latex label, but *not a prior*)

* Re-defining **sampled** into:

 - **fixed**: simply assign it a value
 - **sampled**: change any of prior/ref/latex, etc, to your taste; the rest are inherited
   from the defaults
 - **derived**: mention the parameter and assign nothing to it (you may define a
   min, max or latex label, but *not a prior*; the label is inherited from the defaults)

* Re-defining **derived** into:

 - **fixed**: simply assign it a value
 - **sampled**: define a prior for it (and optionally reference, label, etc.)
 - **derived**: change any of prior/ref/latex, etc, to your taste; the rest are inherited
   from the defaults

As general rules, when trying to redefine anything in a parameter, everything not re-defined
is inherited, except for the prior, which must not be inheritable in order to allow
re-defining sampled parameters into something else.

Custom, multidimensional priors
-------------------------------

The priors described above are
obviously 1-dimensional. You can also define more complicated, multidimensional priors
using a ``prior`` block at the base level (i.e. not indented).

Inside the ``prior`` block, list a pair of priors as ``[name]: [function]``, where the
functions must return **log-priors**.
A prior function can be specified in three different ways:

a) **As a string:** If a string is given, it will be passed to Python's ``eval()`` function. In this case, use Python's ``lambda`` to describe the function, using known parameter names as its arguments. In your string you can access ``scipy.stats`` and  ``numpy`` members under ``stats`` and ``np`` respectively.

b) **As a function:** If your function is more complicated (it takes more than a line of code) and cannot be written as a ``lambda``, you can pass a *function object* directly, whose arguments must be known parameter names. This option is of course not compatible with shell invocation, since a ``yaml`` file can only contain text.

c) **As an import statement:** If your function cannot be written as a ``lambda`` but you  would still like to have your input as a ``yaml`` file, you can code your function into a separate file and use ``import_module`` to load it on the fly.

Let us apply this to the example in section :ref:`in_example`. We will add a
*gaussian ring* prior on both parameters, centred at the origin, with radius 0.5 and sigma of 0.1;
and a simple *gaussian* prior on the sum of both parameters,
centred at 0.5 and with sigma of 0.2.

Using option **(a)** above, we would simply add another block to the input file:

.. code-block:: yaml

   prior:
     ring: "lambda mock_a, mock_b: stats.norm.logpdf(np.sqrt(mock_a**2 + mock_b**2), loc=0.5, scale=0.1)"
     linear: "lambda y: stats.norm.logpdf(mock_a + mock_b, loc=0.5, scale=0.2)"

.. warning::
   Use quotation marks, ``"`` or ``'``, to enclose the prior! Otherwise, the colon of the
   ``lambda`` will be wrongly interpreted by ``yaml``.

We may consider that these priors, or any other, are too complicated for a one-liner,
so we may prefer to code it in a separate file, say ``myprior.py`` in the same folder:

.. code-block:: python

   from scipy.stats import norm
   import numpy as np

   def gaussian_ring(mock_a, mock_b):
       r = np.sqrt(mock_a**2 + mock_b**2)
       return norm.logpdf(r, loc=0.5, scale=0.1)

   def linear_combination(mock_a, mock_b):
       return norm.logpdf(mock_a + mock_b, loc=0.5, scale=0.2)

In this case, which corresponds to option **(c)**, the prior block would look like:

.. code:: yaml

   prior:
     ring: import_module("myprior").gaussian_ring
     linear: import_module("myprior").linear_combination

This way, you can invoke `cobaya` from the command line with the corresponding ``yaml``
file as input, provided that ``myprior.py`` is in the same folder. It is the most
*portable* version, in case you want to pass your sample around, or archive it for future
reference.

CHECK IF DIFFERENT FOLDERS WORK!!!

If you are running `cobaya` within a Python script or a Jupyter notebook, you can
simply pass a dictionary of functions and their names in under
the ``prior`` key of the *information* dicctionary that is passed to ``cobaya.run.run``,
as described in case **(b)**.

Following the example in section :ref:`in_example_script` and using the priors above,
now using case , simply define the ``gaussian_ring`` and
``linear_combination`` functions somewhere early in the script and assign them to the
dictionary containing all the information, before calling ``cobaya.run.run``:

.. code:: python

   info["prior"] = {"ring": gaussian_ring, "linear": linear_combination}



decir que no estan normalised (link a seccion (TODO) sobre normalizar posteriors)

TEST EXAMPLES!!!

YOU MUST STLL SPECIFY THE LIMITS!!!


.. image:: img_prior_ring.svg

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
from collections import OrderedDict as odict
import numpy as np
import numbers
import inspect

# Local
from cobaya.conventions import _prior, _p_dist, _p_ref, _p_label
from cobaya.tools import get_external_function, get_scipy_1d_pdf
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


class Prior():
    """
    Class managing the prior and reference pdf's.
    """
    def __init__(self, parametrisation, info_prior=None):
        """
        Initialises the prior and reference pdf's from the input information.
        """
        sampled_params_info = parametrisation.sampled_params()
        if not sampled_params_info:
            log.warning("No sampled parameters requested! This will fail for non-mock samplers.")
        # pdf: a list of independent components
        # in principle, separable: one per parameter
        self.name = []
        self.pdf = []
        self.ref_pdf = []
        self.lims = []
        for p in sampled_params_info:
            self.name += [p]
            prior = sampled_params_info[p].get(_prior)
            self.pdf += [get_scipy_1d_pdf({p: prior})]
            # Get the reference (1d) pdf
            ref = sampled_params_info[p].get(_p_ref)
            # Cases: number, pdf (something, but not a number), nothing
            if isinstance(ref, numbers.Number):
                self.ref_pdf += [ref]
            elif ref != None:
                self.ref_pdf += [get_scipy_1d_pdf({p: ref})]
            else:
                self.ref_pdf += [np.nan]
            self.lims += [(-np.inf,np.inf)]
            if self.pdf[-1].__class__.__name__ == "rv_frozen":
                self.lims[-1] = self.pdf[-1].interval(1)
            else:
                log.error("No limits defined for parameter '%s'.", param)
                raise HandledException
        # Process the external prior(s):
        self.external = odict()
        for name in (info_prior if info_prior else {}):
            log.debug("Loading external prior '%s' from: '%s'", name, info_prior[name])
            self.external[name] = {"logp": get_external_function(info_prior[name])}
            self.external[name]["argspec"] = inspect.getargspec(self.external[name]["logp"])
            try:
                self.external[name]["indices"] = [list(sampled_params_info).index(p)
                                                  for p in self.external[name]["argspec"].args]
            except ValueError:
                log.error(
                    "The arguments of the external prior '%s' must be known *sampled* parameters. "
                    "Got %r", name, self.external[name]["argspec"].args)
                raise HandledException
            log.warning("External prior '%s' loaded. Mind that it might not be normalised!", name)

    def d(self):
        """
        Returns:
           Dimensionality of the parameter space.
        """
        return len(self.pdf)

    def limits(self):
        """
        Returns:
           An array of limits ``[min,max]`` for the parameters, in the order given by the
           input.

        NB: If an external prior has been defined, the limits given in the 'prior'
        sub-block of that particular parameter's info may not be faithful to the externally
        defined prior.
        """
        return np.array(self.lims.values())

    def sample(self, n=1, external_error=True):
        """
        Generates samples from the prior pdf.

        If an external prior has been defined, it is not possible to sample from the prior
        directly. In that case, if you want to sample from the "default" pdf (i.e.
        ignoring the external prior), set `external_error` to `False`.

        Returns:
          An array of ``n`` samples from the prior, as vectors ``[value of param 1, ...]``.
        """
        if external_error and self.external:
            log.error("It is not possible to sample from an external prior.")
            raise HandledException
        return np.array([pdf.rvs(n) for pdf in self.pdf]).T

    def p(self, x):
        """
        Returns:
           The probability density of the given point or array of points.
        """
        return (np.prod([pdf.pdf(xi) for pdf,xi in zip(self.pdf,x)])
                * np.exp(self.logp_external(x)))

    def logp(self, x):
        """
        Returns:
           The log-probability density of the given point or array of points.
        """
        return (sum([pdf.logpdf(xi) for pdf,xi in zip(self.pdf,x)])
                + self.logp_external(x))

    def logp_external(self, x):
        """Evaluates the logprior using the external prior only."""
        return sum([ext["logp"](*x[ext["indices"]]) for ext in self.external.values()])

    def covmat(self, external_error=True):
        """
        Returns:
           The covariance matrix of the prior.
        """
        if external_error and self.external:
            log.error("It is not possible to get the covariance matrix from an external prior.")
            raise HandledException
        return np.diag([pdf.var() for pdf in self.pdf]).T

    def reference(self, max_tries=np.inf):
        """
        Returns:
          One sample from the ref pdf. For those parameters that do not have a ref pdf
          defined, their value is sampled from the prior.

        NB: The way this function works may be a little dangerous:
        if two parameters have an (external)
        joint prior defined and only one of them has a reference pdf, one should
        sample the other from the joint prior *conditioned* to the first one being
        drawn from the reference. Otherwise, one may end up with a point with null
        prior pdf. In any case, it should work for getting initial reference points,
        e.g. for an MCMC chain.
        """
        if np.nan in self.ref_pdf:
            log.info("Reference values or pdf's for some parameters were not provided. "
                     "Sampling from the prior instead for those parameters.")
        while max_tries:
            max_tries -= 1
            ref_sample = np.array([getattr(ref_pdf, "rvs", lambda: ref_pdf.real)()
                                   for i, ref_pdf in enumerate(self.ref_pdf)])
            where_no_ref = np.isnan(ref_sample)
            if np.any(where_no_ref):
                prior_sample = self.sample(external_error=False)[0]
                ref_sample[where_no_ref] = prior_sample[where_no_ref]
            if self.logp(ref_sample) > -np.inf:
                return ref_sample
        log.error("Couldn't sample from the reference pdf a point with non-"
                  "null prior density after '%d' tries. "
                  "Maybe your prior is improper of your reference pdf is "
                  "null-defined in the domain of the prior.", max_tries)
        raise HandledException

    def reference_covmat(self):
        """
        Returns:
          The standard deviation of the 1d ref pdf's. For those parameters that do not have
          a ref pdf defined, the standard deviation of the prior is taken instead.
        """
        covmat = np.diag([getattr(ref_pdf, "var", lambda: np.nan)()
                          for i, ref_pdf in enumerate(self.ref_pdf)])
        where_no_ref = np.isnan(covmat)
        if np.any(where_no_ref):
            log.warning("Reference pdf not defined or improper for some parameters. "
                        "Using prior's sigma instead for them.")
            covmat[where_no_ref] = self.covmat(external_error=False)[where_no_ref]
        return covmat

    # Python magic for the "with" statement
    def __enter__(self):
        return self
    def __exit__(self, exception_type, exception_value, traceback):
        pass



"""
.. module:: yaml_custom

:Synopsis: Custom YAML loader and dumper
:Author: Jesus Torrado (parts of the code comes from stackoverflow user's)

Customisation of YAML's loaded and dumper:

1. Matches 1e2 as 100 (no need for dot, or sign after e), from http://stackoverflow.com/a/30462009
2. Wrapper to load mappings as OrderedDict (for likelihoods and params), from http://stackoverflow.com/a/21912744

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import yaml
import re
from collections import OrderedDict as odict
import numpy as np

# Exceptions
class InputSyntaxError(Exception):
    """Syntax error in YAML input."""


def yaml_load(text_stream, Loader=yaml.Loader, object_pairs_hook=odict, file_name=None):
        class OrderedLoader(Loader):
            pass
        def construct_mapping(loader, node):
            loader.flatten_mapping(node)
            return object_pairs_hook(loader.construct_pairs(node))
        OrderedLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)
        OrderedLoader.add_implicit_resolver(
            u'tag:yaml.org,2002:float',
            re.compile(u'''^(?:
            [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
            |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
            |\\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
            |[-+]?\\.(?:inf|Inf|INF)
            |\\.(?:nan|NaN|NAN))$''', re.X),
            list(u'-+0123456789.'))
        try:
            return yaml.load(text_stream, OrderedLoader)
        # Redefining the general exception to give more user-friendly information
        except yaml.YAMLError, exception:
            errstr = "Error in your input file "+("'"+file_name+"'" if file_name else "")
            if hasattr(exception, "problem_mark"):
                line   = 1+exception.problem_mark.line
                column = 1+exception.problem_mark.column
                signal = " --> "
                signal_right = "    <---- "
                sep = "|"
                context = 4
                lines = text_stream.split("\n")
                pre = ((("\n"+" "*len(signal)+sep).
                         join([""]+lines[max(line-1-context,0):line-1])))+"\n"
                errorline = (signal+sep+lines[line-1]
                             + signal_right + "column %s"%column)
                post = ((("\n"+" "*len(signal)+sep).
                         join([""]+lines[line+1-1:min(line+1+context-1,len(lines))])))+"\n"
                raise InputSyntaxError(
                    errstr + " at line %d, column %d."%(line, column)+pre+errorline+post+
                    "Maybe inconsistent indentation, '=' instead of ':', "
                    "no space after ':', or a missing ':' on an empty group?")
            else:
                raise InputSyntaxError(errstr)

def yaml_load_file(input_file):
    """Wrapper to load a yaml file."""
    with open(input_file,"r") as file:
        lines = "".join(file.readlines())
    return yaml_load(lines, file_name=input_file)

def yaml_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
    class OrderedDumper(Dumper):
        pass
    # Dump OrderedDict's as plain dictionaries, but keeping the order
    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items())
    OrderedDumper.add_representer(odict, _dict_representer)
    # Dump tuples as ymal "sequences"
    def _tuple_representer(dumper, data):
        return dumper.represent_sequence(
            yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG, list(data))
    OrderedDumper.add_representer(tuple, _tuple_representer)
    # Numpy arrays and numbers
    def _numpy_array_representer(dumper, data):
        return dumper.represent_sequence(
            yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG, data.tolist())
    OrderedDumper.add_representer(np.ndarray, _numpy_array_representer)
    def _numpy_int_representer(dumper, data):
        return dumper.represent_int(data)
    OrderedDumper.add_representer(np.int64, _numpy_int_representer)
    def _numpy_float_representer(dumper, data):
        return dumper.represent_float(data)
    OrderedDumper.add_representer(np.float64, _numpy_float_representer)
    # Dump!
    return yaml.dump(data, stream, OrderedDumper, **kwds)



"""
.. module:: tools

:Synopsis: General tools
:Author: Jesus Torrado

"""

# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import os
import sys
from collections import OrderedDict as odict
from copy import deepcopy
from importlib import import_module
import numpy as np
import inspect
import scipy.stats as stats

# Local
from cobaya.conventions import package, subfolders, _likelihood
from cobaya.conventions import _p_label, _p_dist, _prior, _debug
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)

_path_to_installation = None

def set_path_to_installation(path):
    global _path_to_installation
    _path_to_installation = path

def get_path_to_installation():
    return _path_to_installation

def get_folder(name, kind, sep="/", absolute="True"):
    """
    Gets folder, relative to the package source, of a likelihood, theory or sampler.
    """
    pre = os.path.dirname(__file__)+sep if absolute else ""+(sep if sep=="." else "")
    return pre + subfolders[kind] + sep + name

def get_class(name, kind=_likelihood):
    """
    Retrieves the requested likelihood (default) or theory class.

    ``info`` must be a dictionary of the kind ``{[class_name]: [options]}``.

    Raises ``ImportError`` if class not found in the appropriate place in the source tree.
    """
    class_folder = get_folder(name, kind, sep=".", absolute=False)
    try:
        return getattr(import_module(class_folder, package=package), name)
    except:
        if sys.exc_info()[0] == ImportError and str(sys.exc_info()[1]).endswith(name):
            log.error("%s '%s' not found (wrong capitalisation?)", kind.capitalize(), name)
            raise HandledException
        else:
            log.error("There was a problem when importing %s '%s':", kind, name)
            raise sys.exc_info()[1]

def get_external_function(string_or_function):
    """
    Processes an external prior or likelihood, given as a string or a function.

    If the input is a string, it must be evaluable to a function. It can contain import
    statements using :module:`importlib`'s ``import_module``, e.g.
    ``import_module("my_file").my_function``. You can access :module:`scipy.stats` and
    :module:`numpy` members under the handles ``stats`` and ``np`` respectively.

    Returns the function.
    """
    if isinstance(string_or_function, basestring):
        try:
            function = eval(string_or_function)
        except Exception, e:
            log.error("Failed to load external function: '%r'", e)
            raise HandledException
    else:
        function = string_or_function
    if not callable(function):
        log.error("The external function provided is not an actual function.")
        raise HandledException
    return function

def make_header(kind, module):
    """Created a header for a particular module of a particular kind."""
    return ("="*80).join(["", "\n %s : %s \n"%(kind.title(), module), "\n"])

def ensure_latex(string):
    """Inserts $'s at the beginning and end of the string, if necessary."""
    if string.strip()[0] != r"$":
        string = r"$"+string
    if string.strip()[-1] != r"$":
        string = string+r"$"
    return string

def ensure_nolatex(string):
    """Removes $'s at the beginning and end of the string, if necessary."""
    return string.strip().lstrip("$").rstrip("$")

def get_scipy_1d_pdf(info):
    """Generates 1d priors from scipy's pdf's from input info."""
    param = info.keys()[0]
    info2 = deepcopy(info[param])
    # What distribution?
    try:
        dist = info2.pop(_p_dist).lower()
    # Not specified: uniform by default
    except KeyError:
        dist = "uniform"
    # Number: uniform with 0 width
    except AttributeError:
        dist = "uniform"
        info2 = {"loc":info2, "scale":0}
    try:
        pdf_dist = getattr(import_module("scipy.stats", dist), dist)
    except AttributeError:
        log.error("Error creating the prior for parameter '%s': "
                  "The distribution '%s' is unknown to 'scipy.stats'. "
                  "Check the list of allowed possibilities in the docs.", param, dist)
        raise HandledException
    # Recover loc,scale from min,max
    # For coherence with scipy.stats, defaults are min,max=0,1
    if "min" in info2 or "max" in info2:
        if "loc" in info2 or "scale" in info2:
            log.error("You cannot use the 'loc/scale' convention and the 'min/max' "
                      "convention at the same time. Either use one or the other.")
            raise HandledException
        try:
            mini = np.float(info2.pop("min"))
        except KeyError:
            mini = 0
        try:
            maxi = np.float(info2.pop("max"))
        except KeyError:
            maxi = 1
        info2["loc"] = mini
        info2["scale"] = maxi-mini
    # Generate and return the frozen distribution
    try:
        return pdf_dist(**info2)
    except TypeError as tp:
        log.error(
            "'scipy.stats' produced an error: <<%r>>. "
            "This probably means that the distribution '%s' "
            "does not recognise the parameter mentioned in the 'scipy' error above.",
            tp.message, dist)
        raise HandledException
